<?php

// if display_errors = on, errors will result in a 200 status, so setting the 
// status to 500 ahead of time, then resetting to 200 on success
// see: https://bugs.php.net/bug.php?id=50921
header('HTTP/1.0 500 Internal Server Error');

/**
 * RESTful Query Params:
 *   - start
 *   - end
 *      - seconds since epoch
 *         - i.e. "X:datetime"
 *      - seconds ago
 *         - i.e. "Y:ago"
 *   - group
 *      - i.e. 86400 (1 day), week, month, year
 *   - data
 *      - reading:agg:unit:decimals
 *      - i.e. "outTemp,outDew:mean::1,bar::mb"
 *   - options
 *      - defaultDecimals
 *
 * POST JSON structure:
 *   {
 *      start : { value : X, type : "datetime|ago" },
 *      end : { ... },
 *      group : { value : X, type : "period" }
 *              { value : |"all"|"month|year", type : "period|calendar" },
 *      data : [
 *         {
 *            reading : "dateTime|outTemp|outDew|outHum|rain|rainRate|wind|windDir|windGust|windGustDir|bar|heatIndex|windChill|inTemp|inHum", 
 *            agg : "avg|mean|min|max|sum",
 *            unit : "f|c| mb|inHg| in|mm|cm| inHr|mmHr|cmHr| mph|mps|kph|knot",
 *            decimals : X
 *         }
 *     ]
 *   }
 *
 */
require_once("include/JsonConfig.class.php");
require_once("include/AggregateParameterParser.class.php");
require_once("include/AggregateQuery.class.php");
require_once("include/PDOConnectionFactory.class.php");
require_once 'include/TableEntity.class.php';

$config = JsonConfig::getInstance();

$parser = new AggregateParameterParser($_REQUEST, $config);

$spec = $parser->parse();

// FIXME for now assuming all entities are tables
$entity = new TableEntity($spec->entityId, $config);

//var_dump($spec);

// TODO need to support calculated fields (i.e. actual rainRate)
$query = new AggregateQuery($spec);

$db_config = $config['dataSource'][$spec->entityConfig['dataSource']];

header( "X-Meso-Query: ". str_replace("\n", " ", $query->sql) );

$db = PDOConnectionFactory::openConnection($db_config);

$_start = microtime(true);

try {
    $data = $db->query( $query->sql );
} catch(PDOException $e) { // XXX a quick and dirty fix for automatically creating the table if it doesn't yet exist
    // table doesn't exist eror code: 42S02
    if($e->getCode() == '42S02') {
        // attempt to create the table and retry
        $entity->createTable();
        $data = $db->query( $query->sql );
    } else {
        // re-throw all other errors encountered
        throw $e;
    }
}

header( "X-Meso-Query-Time: ". (microtime(true) - $_start) );

$data->setFetchMode(PDO::FETCH_NUM);

// success! reset to 200 status
header('HTTP/1.0 200 OK');

echo '[';

$i = 1;
while( $row = $data->fetch() ) {
    if( $i != 1 ) echo ",";
    $row = array_map( function($column) {
        return $column === NULL ? "null" : $column;
    }, $row);
    echo "[" . join( ",", $row ) . "]";
    $i++;
}

echo ']';

?>
